#include <QSettings>
#include <QDir>
#include <QMessageBox>
#include <QFile>
#include <QFileDialog>
#include <QApplication>
#include <QPixmap>
#include <QRegExp>
#include <QVector>
#include <QTextStream>
#include <QDrag>
#include <QProcess>
#include <QUrl>
#include <QDesktopServices>
#include <QDragEnterEvent>
#include <QDropEvent>

#include "waveread16.h"
#include "hdr_wav.h"

#if defined(Q_OS_WIN32)
#include <windows.h>
#endif

#include "wmain.h"
#include "moc_wmain.cpp"

#define BUFSIZE (1048576*2) //2mb r/w buffer

QString WMain::fnPrefix(const QString &s_)
{
    QFileInfo fi(s_);
    return fi.completeBaseName()+": ";
}

QString WMain::bareFile(const QString &s_)
{
    QFileInfo fi(s_);
    return fi.completeBaseName();
}

QString WMain::toFname(const QString &_s)
{
    QString s=_s;
    
    s.replace("/","-");
    s.replace("\\","_");
    s.replace(":","_");
    s.replace(";",";");
    s.replace("&"," and ");
    s.replace("+","_");
    //s.replace("#","sharp");
    s.replace("\"","");
    s.replace("`","'");
    s.replace("?","");
    //s.replace("!","");
    //s.replace("@"," at ");
    s.replace("*","-");
    
    s = s.simplified();
    s = s.trimmed();
    
    return s;
}

WMain::WMain()
{
    QSettings ss;

    setupUi(this); 
    
#ifdef Q_WS_MACX
    btnExtract->setMinimumSize(0,32);
    btnStop->setMinimumSize(0,32);
    actionRemove->setShortcut(QKeySequence("Backspace"));
#endif       
    
    destPath->setText(ss.value("destPath","").toString());
    subArtist->setChecked(ss.value("subfolderArtists",true).toBool());
    subAlbum->setChecked(ss.value("subfolderAlbums",true).toBool());
    cbOverwrite->setChecked(ss.value("overwriteExisting",false).toBool());    
    trackFormat->setEditText(ss.value("trackFormat","%1 %2 - %3").toString());
    on_trackFormat_textChanged("");
    
    cuelist->setFocus();
}

WMain::~WMain()
{
    QSettings ss;
    
    ss.setValue("destPath",destPath->text());
    ss.setValue("subfolderArtists",subArtist->isChecked());
    ss.setValue("subfolderAlbums",subAlbum->isChecked());    
    ss.setValue("overwriteExisting",cbOverwrite->isChecked());    
    ss.setValue("trackFormat",trackFormat->currentText());        
}


void WMain::on_actionStartExtraction_triggered()
{
    QDir dir(destPath->text());
    if (destPath->text().trimmed().isEmpty()
        ||(!dir.exists()))
    {
        QMessageBox::information(this,
                                 tr("Information"),
                                 tr("Destination path is not valid (does not exist).\n"
                                 "Please select the existing directory."));
        return;
    }
    
    if (!cuelist->count())
    {
        QMessageBox::information(this,
                                 tr("Information"),
                                 tr("CUE list is empty - nothing to extract."));
        return; 
    }
    
    QWidget *widgets[]={
        destPath,subArtist,subAlbum,cbOverwrite,trackFormat,
        btnExtract,cuelist,btnBrowse,0};
    
    int t;
    for(t=0;widgets[t];t++)
        widgets[t]->setEnabled(false);
    
    actionAdd->setEnabled(false);
    actionRemove->setEnabled(false);    
    actionStartExtraction->setEnabled(false);
    actionQuit->setEnabled(false);    
    
    btnStop->setEnabled(true);
    actionStop->setEnabled(true);
    
    stopWanted=false;
      
    doExtract();
    
    actionStop->setEnabled(false);    
    btnStop->setEnabled(false);    
    
    for(t=0;widgets[t];t++)
        widgets[t]->setEnabled(true);        
    
    actionAdd->setEnabled(true);
    actionRemove->setEnabled(true);    
    actionStartExtraction->setEnabled(true);
    actionQuit->setEnabled(true);    
    stopWanted=false;
}


void WMain::on_actionStop_triggered()
{    
    stopWanted=true;
    QListWidgetItem *p=new QListWidgetItem(QIcon(":/stopme.png"),"Stop!");
    loglist->addItem(p);
    {if (loglist->count()) loglist->setCurrentRow(loglist->count()-1);}
    loglist->scrollToItem(p);
    qApp->processEvents();        
}


void WMain::on_btnBrowse_clicked()
{
    QString str = 
            QFileDialog::getExistingDirectory(
                    this,
                    tr("Select a destination path for extraction"),
                    destPath->text());
    
    if (str.isEmpty()) return;
    
    QDir dir(str);
    dir.makeAbsolute();
    str = QDir::convertSeparators(dir.absolutePath());
    destPath->setText(str);    
}

void WMain::on_destPath_editingFinished()
{
    QString str=destPath->text();
    str = str.trimmed();
    if (str.isEmpty())
        destPath->setText(str);
    
    QDir dir(str);
    if (dir.exists())
    {
        dir.makeAbsolute();
        str = QDir::convertSeparators(dir.absolutePath());
        destPath->setText(str);
    }
}


void WMain::addFileList(QStringList lst)
{
    QIcon px(":/block.png");
    
    cuelist->setSortingEnabled(false);    
    
    QMap<QString,bool> present;
    
    for(int k=0;k<cuelist->count();k++)
    {
        QListWidgetItem *item=cuelist->item(k);
        present[item->text()]=true;
    }          
    
    QStringList::iterator i;
    for(i=lst.begin();i!=lst.end();i++)
    {
        QFileInfo fi((*i));     
        QString fp = QDir::convertSeparators(fi.absoluteFilePath());
        if (!fp.toLower().trimmed().endsWith(".cue")) continue;
        if (fi.exists()&&fi.isReadable()&&(fi.size()<=100000)&&(!present[fp]))
        {           
            QListWidgetItem *item = new QListWidgetItem(fp,cuelist);
            item->setIcon(px);
            present[fp]=true;
        }       
    }    
    cuelist->setSortingEnabled(true);
    cuelist->sortItems();
}
        
void WMain::on_actionAdd_triggered()
{
    QStringList lst = 
            QFileDialog::getOpenFileNames(
                    this,
                    tr("Add CUE files"), 
                    QString::null,
                    tr("CUE files (*.cue)"));

    if (!lst.isEmpty()) addFileList(lst);
}

void WMain::selectCue(const QString &s_)
{
    QString s = QDir::convertSeparators(s_);
    for(int k=0;k<cuelist->count();k++)
    {
        QListWidgetItem *item = cuelist->item(k);

        if (item->text()==s)
        {
            item->setIcon(QIcon(":/logerror.png"));
            item->setSelected(true);
        }
    } 
    qApp->processEvents();
}

void WMain::on_actionRemove_triggered()
{
    QList<QListWidgetItem *> lst;
    QList<QListWidgetItem *>::iterator i;
    
    for(int k=0;k<cuelist->count();k++)
    {
        QListWidgetItem *item = cuelist->item(k);

        if (item->isSelected())
            lst.append(item);
    }

    for(int k=0;k<cuelist->count();k++)
    {
        QListWidgetItem *item = cuelist->item(k);
        item->setSelected(false);
    }

    for(i=lst.begin();i!=lst.end();i++)
        delete (*i);    
}


void WMain::on_actionQuit_triggered()
{
    close();
}

void WMain::on_btnClearLog_clicked()
{
    loglist->clear();
}

void WMain::lognormal(const QString &s)
{
    QListWidgetItem *p=new QListWidgetItem(QIcon(":/lognormal.png"),s);
    loglist->addItem(p);
    {if (loglist->count()) loglist->setCurrentRow(loglist->count()-1);}
    loglist->scrollToItem(p);
    qApp->processEvents();
}

void WMain::logerror(const QString &s)
{
    QListWidgetItem *p=new QListWidgetItem(QIcon(":/logerror.png"),s);
    loglist->addItem(p);
    {if (loglist->count()) loglist->setCurrentRow(loglist->count()-1);}
    loglist->scrollToItem(p);
    qApp->processEvents();
}

int WMain::unpackCue(const QString &fname,bool checkOnly)
{
    QString userterm=tr("User termination");

#define CHECKSTOP() \
    {qApp->processEvents();\
     if (stopWanted) {logerror(userterm);selectCue(fname);return 0;}\
    }
    
    if (checkOnly)
    {
        //lognormal(QString("Checking %1").arg(fname));
    }else
    {
        lognormal(tr("Extracting tracks from %1").arg(QDir::convertSeparators(fname)));
    }
    
    QFile f(fname);
    
    if (!f.exists())
    {
        logerror(tr("File not found (%1)").arg(QDir::convertSeparators(fname)));
        selectCue(fname);
        return 0;
    }
    
    if (!f.open(QIODevice::ReadOnly))
    {
        logerror(tr("Cannot open %1 for reading").arg(QDir::convertSeparators(fname)));
        selectCue(fname);
        return 0;
    }
    if (f.size()>100000)
    {
        logerror(tr("File %1 is too large").arg(QDir::convertSeparators(fname)));
        selectCue(fname);
        return 0;
    }
    
    QTextStream tf(&f);
    
    QString performer=tr("Unknown Artist");
    QString title=tr("Unknown Title");
    QString wavefile=fname;

    wavefile.replace(".cue",".wav");
    wavefile.replace(".CUE",".WAV");    
    wavefile.replace(".Cue",".Wav");        
    wavefile.replace(".cue",".wav",Qt::CaseInsensitive);            
    
    QString s;
    
    QRegExp rx_performer("^PERFORMER\\s*\"\\s*([^\"]*)\"$",Qt::CaseInsensitive);
    QRegExp rx_title    ("^TITLE\\s*\"\\s*([^\"]*)\"$",Qt::CaseInsensitive);
    QRegExp rx_file     ("^FILE\\s*\"\\s*([^\"]*)\"\\s*WAVE$",Qt::CaseInsensitive);
    QRegExp rx_track    ("^TRACK\\s*([0-9]+)\\s*AUDIO$",Qt::CaseInsensitive);
    QRegExp rx_index00  ("^INDEX\\s*00\\s*([0-9]+):([0-9]+):([0-9]+)$",Qt::CaseInsensitive);
    QRegExp rx_index01  ("^INDEX\\s*01\\s*([0-9]+):([0-9]+):([0-9]+)$",Qt::CaseInsensitive);
    
    QVector<QString> xtrackno;
    QVector<QString> xperf;
    QVector<QString> xtitle;
    QVector<qint64> xbegin;
    QVector<qint64> xend;
    
    for(;;)
    {
        CHECKSTOP();    
        s=tf.readLine();
        if (s.isNull()) break;
        s=s.trimmed();
        if (s.isEmpty()) continue;
        
        if (s.startsWith("CATALOG"))
        { 
            continue;
        }else
        if (s.startsWith("FLAGS"))
        { 
            continue;
        }else
        if (s.startsWith("REM "))
        { 
            continue;
        }else
        if (s.startsWith("ISRC"))
        { 
            continue;
        }else
        if (rx_performer.exactMatch(s))
        {
            if (xperf.size()==0)
                performer=rx_performer.cap(1);
            else
                xperf[xperf.size()-1] = rx_performer.cap(1);
        }else if (rx_title.exactMatch(s))
        {
            if (xtitle.size()==0)
                title=rx_title.cap(1);
            else
                xtitle[xtitle.size()-1] = rx_title.cap(1);
        }else if (rx_file.exactMatch(s))
        {
            if (xtitle.size()!=0)
            {
                logerror(fnPrefix(fname)+
                         tr("'FILE' found after track enlisting"));
                selectCue(fname);
                return 0;
            }else
            {
                wavefile = rx_file.cap(1);
                wavefile.replace("\\","/");
                QFileInfo wfi(wavefile);
                QFileInfo fi(fname);
                QFile ff1(fi.absolutePath()+"/"+wfi.fileName());
                if (!ff1.exists())
                {
                    logerror(QDir::convertSeparators(fname)+
                             ": "+tr("Cannot find %1")
                             .arg(QDir::convertSeparators(ff1.fileName())));
                    selectCue(fname);               
                    return 0;
                }
                wavefile = ff1.fileName();
            }
        }else if (rx_track.exactMatch(s))
        {
            if (xbegin.size()>0)
            {
                if (xbegin[xbegin.size()-1]==-1)
                {
                    logerror(QDir::convertSeparators(fname)
                             +": "+tr("Track without an index found"));
                    selectCue(fname);               
                    return 0;
                }
            }
            if (xend.size()>1)
            {
                if (xend[xend.size()-2]==-1)
                    xend[xend.size()-2] = xbegin[xbegin.size()-1];
            }
            
            xperf.append(performer);
            xtrackno.append(rx_track.cap(1));
            xtitle.append(tr("Track %1").arg(rx_track.cap(1)));
            xbegin.append(-1);
            xend.append(-1);
        }else if (rx_index01.exactMatch(s))
        {
            qint64 st = 
                    rx_index01.cap(1).toLongLong()*100*60 +
                    rx_index01.cap(2).toLongLong()*100 +
                    rx_index01.cap(3).toLongLong()*1;
            
            if (xtitle.size()==0)
            {
                logerror(QDir::convertSeparators(fname)
                         +": "+tr("Index without a track found"));
                selectCue(fname);                   
                return 0;
            }
            
            if (xbegin[xbegin.size()-1]!=-1)
            {
                logerror(QDir::convertSeparators(fname)
                         +": "+tr("Double index found"));
                selectCue(fname);                   
                return 0;
            }

            if ((xend.size()>1)&&(xend[xend.size()-2]==-1))
                xend[xend.size()-2] = st;
            
            xbegin[xbegin.size()-1] = st;
        }else if (rx_index00.exactMatch(s))
        {
            qint64 st = 
                    rx_index00.cap(1).toLongLong()*100*60 +
                    rx_index00.cap(2).toLongLong()*100 +
                    rx_index00.cap(3).toLongLong()*1;
            
            if (xend.size()>1)
            {
                xend[xend.size()-2] = st;
            }
        }else
        {
            logerror(QDir::convertSeparators(fname)
                     +": "+tr("Invalid line found"));            
            logerror(QString("    ")+s);
            selectCue(fname);                       
            return 0;
        }
    }
    
    QString txt_various = tr("Various (No information)");

    if ((performer==tr("Unknown Artist"))&&(title==tr("Unknown Title")))
    {
        performer=txt_various;
        QFileInfo fi(fname);
        title = fi.completeBaseName();
    }
    
    CHECKSTOP();    
    WaveRead16 rd;    
    if (rd.open(wavefile)<=0)
    {
        logerror(QDir::convertSeparators(fname)
                 +": "+tr("Failed to open %1").arg(wavefile));     
        selectCue(fname);                               
        return 0;
    }
    
    int t;
    
    if (checkOnly) return 1;
    
    QDir cdir(".");
    
    QString path = destPath->text()+"/";
    
    performer = toFname(performer);
    title = toFname(title);
    
    CHECKSTOP();
    if (subArtist->isChecked())
    {
        if (!cdir.exists(path+performer))
        {
            if (!cdir.mkdir(path+performer))
            {
                logerror(QDir::convertSeparators(fname)
                         +": "+tr("Cannot create directory %1")
                         .arg(QDir::convertSeparators(path+performer)));    
                selectCue(fname);
                return 0;
            }
        }
        path+=performer+"/";
    }
    
    CHECKSTOP();
    if ((subAlbum->isChecked())||(performer==txt_various))
    {
        if ((!subArtist->isChecked())&&(performer!=txt_various))
            title = performer + " - " + title;

        if (!cdir.exists(path+title))
        {
            if (!cdir.mkdir(path+title))
            {
                logerror(fnPrefix(fname)+
                         tr("Cannot create directory %1")
                         .arg(QDir::convertSeparators(path+title)));  
                selectCue(fname);           
                return 0;
            }
        }
        path+=title+"/";
    }
    
    for(t=0;t<(int)xtitle.size();t++)
    {
        QString strf = trackFormat->currentText().trimmed();    
        strf.replace("%1",xtrackno[t]);
        strf.replace("%2",xperf[t]);    
        strf.replace("%3",xtitle[t]);   
        strf.replace("%4",title);
        QString wrfile = path+
                         toFname(strf.trimmed()+".wav");
        
        CHECKSTOP();    
        QFile wf(wrfile);
        
        if (!cbOverwrite->isChecked())
        {
            if (wf.exists())
            {
                lognormal(fnPrefix(fname)+
                          tr("Skipping existing file %1")
                          .arg(QDir::convertSeparators(wrfile)));
                continue;
            }
        }else
        {
            if (wf.exists())
            {
                lognormal(fnPrefix(fname)+
                          tr("Overwriting existing file %1")
                          .arg(QDir::convertSeparators(wrfile)));
            }       
        }
        
        xbegin[t] = xbegin[t] * ((qint64)rd.rate) / ((qint64)100);
        
        if ((xend[t]==-1)&&(t==(xend.size()-1))) 
        {
            xend[t]=rd.sz;
        }else
            xend[t] = xend[t] * ((qint64)rd.rate) / ((qint64)100);
        
        qint64 sz=xend[t]-xbegin[t];
        
        if (sz<=0) 
        {
            lognormal(fnPrefix(fname)+
                    tr("Warning in file %1").arg(QDir::convertSeparators(wrfile)));
            lognormal(fnPrefix(fname)+
                    tr("No track data, size=%1").arg((int)sz));
            fflush(stdout);
            sz=0;
            continue;
        }
        
        CHECKSTOP();    
        if (!wf.open(QIODevice::WriteOnly))
        {
            logerror(fnPrefix(fname)+
                    tr("Cannot write file %1").arg(QDir::convertSeparators(wrfile)));          
            selectCue(fname);
            return 0;
        }
        
        CHECKSTOP();
        
        lognormal(fnPrefix(fname)+
                  tr("Writing %1...").arg(QDir::convertSeparators(wrfile)));
        
        char hdr[0x2C];
        
        memcpy(hdr,stub_wav_data,0x2C);
        
#define I32AT(hdr,a) (*(int32_LE *)(&hdr[a]))
#define I16AT(hdr,a) (*(int16_LE *)(&hdr[a]))  
        
        
        I32AT(hdr,0x18) = rd.rate;
        I16AT(hdr,0x16) = rd.nch;
        I32AT(hdr,0x28) = sz*rd.nch*2;
        
        wf.write((char*)hdr,0x2C);
        
        static short buf[BUFSIZE*2];
        
        qint64 offs = xbegin[t];
        
        CHECKSTOP();
        while(sz)
        {
            qint64 chk;
            
            memset(buf,0,sizeof(buf));
            
            if (sz>=BUFSIZE)
                chk = rd.read(offs,BUFSIZE,buf);
            else
                chk = rd.read(offs,sz,buf);
            
            if (wf.write((char *)buf,chk*sizeof(short)*rd.nch) !=
                (chk*sizeof(short)*rd.nch))
            {
                logerror(fnPrefix(fname)+
                         tr("Error writing %1 (check disk space?)")
                         .arg(QDir::convertSeparators(wrfile)));                
                selectCue(fname);
                return 0;
            }
                    
            qApp->processEvents();
            if (stopWanted)
            {
                wf.close();
                if (!wf.remove())
                {
                    logerror(fnPrefix(fname)+
                             tr("Cannot delete incomplete file %1")
                             .arg(QDir::convertSeparators(wrfile)));
                }
                CHECKSTOP();
            }
            
            sz-=chk;
            offs+=chk;
        }
                
        int fsize=wf.pos();
        wf.seek(0);
        I32AT(hdr,0x04) = fsize - 8;
        wf.write((char*)hdr,0x2C);
        
        wf.seek(fsize);
        wf.close();
    }
    
    lognormal(tr("%1: Extraction successful")
              .arg(QDir::convertSeparators(fname)));
    
    return 1;
}

void WMain::doExtract()
{
    cuelist->sortItems();

    for(int k=0;k<cuelist->count();k++)
        cuelist->item(k)->setSelected(false);
    
    bool checkFailed=false;
    
    lognormal(tr("Extraction started")+" ----------------");
    lognormal(tr("Checking CUE files..."));    
    
    QIcon px(":/block.png");

    for(int k=0;k<cuelist->count();k++)
    {
        QListWidgetItem *item=cuelist->item(k);

        item->setIcon(px);
        item->setSelected(false);       
    } 

    for(int k=0;k<cuelist->count();k++)
    {
        QListWidgetItem *item=cuelist->item(k);

        if (unpackCue(item->text(),true)<=0)
            checkFailed=true;
    }
    
    if (checkFailed)
    {
        logerror(tr("Extraction failed")+" ----------------"); 
        QMessageBox::information(this,
                                 tr("Information"),
                                 tr("CUE files contained errors or incompatible lines.\n"
                                 "Review CUE entries marked with the error icon.\n\n"
                                 "(We currently do not support all kind of CUE\n"
                                 "files, but only a subset of CUE/WAV format)"));
        return;
    }
    lognormal(tr("Extracting files..."));    
 
    
    QList<QListWidgetItem *> lst;
    QList<QListWidgetItem *>::iterator i;

    for(int k=0;k<cuelist->count();k++)
    {
        QListWidgetItem *item=cuelist->item(k);
        lst.append(item);
    }

    for(i=lst.begin();i!=lst.end();i++)
    {
        if (unpackCue((*i)->text(),false)<=0)
        {
            if (stopWanted) return;
            logerror(tr("Extraction failed")+" ----------------");     
            QMessageBox::information(this,
                                     tr("Information"),
                                     tr("An error occured during extraction.\n"
                                     "Review CUE entry marked with the error icon.\n\n"
                                     "(We currently do not support all kind of CUE\n"
                                     "files, but only a subset of CUE/WAV format)"));
            return;
        }
        delete (*i);
    }
    lognormal(tr("Extraction completed")+" ----------------"); 
}


void WMain::on_trackFormat_textChanged( const QString & )
{
    QString strf=trackFormat->currentText().trimmed();
    strf.replace("%1","01");
    strf.replace("%2",tr("Artist"));        
    strf.replace("%3",tr("Track Name"));    
    strf.replace("%4",tr("Album"));
    strf = strf.trimmed();
    strf+=".wav";
    trackFormatLabel->setText(strf);
}

void WMain::dragEnterEvent(QDragEnterEvent* event)
{
    if (actionStop->isEnabled()) return;
    if (event->mimeData() && event->mimeData()->hasUrls()) event->accept();else event->ignore();
}

void WMain::dropEvent(QDropEvent* event)
{
    if (actionStop->isEnabled()) return;
    if (!(event->mimeData() && event->mimeData()->hasUrls())) return;

    QStringList lst;
    QList<QUrl> urls=event->mimeData()->urls();

    for(int k=0;k<urls.size();k++)
    {
        QString f=urls[k].toLocalFile();

        if (!f.isEmpty()) lst.append(f);
    }


       
    QStringList::iterator i;
    QStringList lstx;
    for(i=lst.begin();i!=lst.end();i++)
    {
        QString f=(*i);
        if (!f.toLower().trimmed().endsWith(".cue")) continue;
        lstx.append(f);
    }        
    addFileList(lstx);
    event->accept();
}


void WMain::on_actionAbout_triggered()
{
    QMessageBox msg(QMessageBox::Information, 
                    "About...",
                    "",
                    QMessageBox::Yes | QMessageBox::Close, this);

    if (!this->windowIcon().isNull()) msg.setWindowIcon(this->windowIcon());
        
    msg.button(QMessageBox::Yes)->setText(tr("Home Page"));
    msg.setDefaultButton(QMessageBox::Close);
    msg.setText(
            this->windowTitle()+"\n"
            "Copyright (C)2005-2009 by George Yohng\n\n"+
            tr("This is an individual track wave extraction utility\n"
            "for CUE/WAV sets."));
    
    if (msg.exec()==QMessageBox::Yes)
    {   
        QDesktopServices::openUrl(QUrl("http://www.yohng.com/software/cue-wav-extract"));
    }
}
